;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

;;;
;;;information browser has two parts. A header display
;;; details important information about the object. A parameter
;;; display allows the object's paramaters to be edited. To setup
;;; a parameter, type its new value in the associated buffer and hit
;;; return. To unset a parameter, type return in an empty buffer.
;;; To inspect all slots, click on the All checkbox.


(defclass info-browser (browser gray-3d-framed-view-mixin)
  ()
  (:default-initargs
    :window-type :document
    :view-size (xy 200 200)))

(defun info-browser? (object)
  (typep object 'info-browser))

(defmethod set-browser-title ((browser info-browser) &optional object)
  (ccl:set-window-title
   browser
   (format nil "~A Info" 
           (object-namestring (or object (object-view-object browser))))))

(defmethod update-window-for-changed-object ((browser info-browser)
                                              object)
  (when (eq object (object-view-object browser))
    (set-browser-title browser object)))

(defmethod make-browser ((class (eql 'info-browser)) object &rest args)
  (apply #'make-instance class :object object args))

(defmethod initialize-instance :after ((browser info-browser) 
                                       &rest args)
  (declare (ignore args))
  (let* ((object (slot-value browser 'object))
         (name? (object-name object))
         (entries '())
         (bold (bold-face *capella-font*))
         (editx  80)
         (readonly (logior +entry-normal-mode+ +entry-readonly+))
         height table1 table2 allbox params)
    
    ;; table1 holds header info in specialized format.
    (when name?
      (push (make-table-entry :id ':name :label "Name:"
                              :mode (if (rename-ok? object) 
                                      +entry-normal-mode+
                                      readonly)
                              :editor (object-namestring object))
            entries))
    (unless (systemp object)
      (push (make-table-entry :id 1 :label "Position:" :mode readonly
                              :editor (or (object-positionstring object) 
                                          "???"))
            entries))
    (push (make-table-entry :id 2 :label "Status:" :mode readonly
                            :editor (status-string (object-flags object)))
          entries)
    (when (typep object 'container)
      (push (make-table-entry :id 3 :label "Subobjects:" :mode readonly
                              :editor (if (typep object 'container)
                                        (object-count object) 0))
            entries))
    (setf table1 
          (make-instance 'table
            :view-nick-name ':header-table
            :label-args (list 'table-label 
                              :view-font bold)
            :edit-x editx
            :view-size (xy (- (view-width browser) 20) 0)
            :view-position (xy 8 8)
            :entries (nreverse entries)))
    (ccl:add-subviews browser table1)

    ;; add slot table header line 
    (setf height (+ +border+ (view-height table1) 10))
    (setf params (make-instance 'label 
                   :view-position (xy 8 height)
                   :view-size (xy 75 14)
                   :title "Parameters:"
                   :view-font bold))
    ;(setf allbox (make-instance 'check-box
    ;               :view-position (xy 100 height)
    ;               :view-size (xy 60 14)
    ;               :title "Inspect"
    ;               :action #'(lambda (box) 
    ;                          (inspect-slots
    ;                            (ccl:view-named ':slot-table
    ;                                            (view-browser box))
    ;                            (toggle-selected? box)))
    ;              :view-nick-name ':all))
    (setf allbox (make-instance 'button
                   :view-position (xy (+ editx 4) height)
                   :view-size (string-extent "Inspect" *capella-font*
                                             (xy 4 2))
                   :title "Inspect"
                   :action 
                   #'(lambda (b) 
                       (inspect (object-view-object (view-browser b))))
                  :view-nick-name ':all))

    (ccl:add-subviews browser params allbox
                      (make-instance 'invisible-default-button
                        :action #'(lambda (b) 
                                    (info-browser-action 
                                     (view-browser b)))
                        :view-nick-name ':ok))
    ;; table2 holds parameter value editors.
    (incf height 20)
    (setf table2
          (make-instance 'slot-table :object object
                         :block '(elements flags id pathname)
                         :view-position (xy 8 height)
                         :view-size (xy (- (view-width browser)
                                           20) 
                                        0)
                         :edit-x editx))
    (ccl:add-subviews browser table2)
    (ccl:set-view-size browser 
                       (xy (view-width browser) 
                           (+ height
                              (view-height table2)  8)))
    (when name?
      (select-table-editor table1 ':name) )
    (set-browser-title browser)))

(defun info-browser-action (browser)
  ;; invoked by #\Return in a table buffer in an info brower. finds the
  ;; fentry from the buffer and perform the appropriate edit operation.
  ;; name and comment are currently handled as special cases. this is 
  ;; unfortunate,  table should be fixed to handle this automaitcally
  (let ((object (object-view-object browser))
        (editor (ccl:current-key-handler browser))
        (table1 (ccl:view-named ':header-table browser))
        (table2 (ccl:view-named ':slot-table browser))
        entry text)
    (setf text (textbuf-text editor)) 
    (setf entry (or (table-find editor table1 
                                      :key #'table-entry-editor)
                    (table-find editor table2 
                                      :key #'table-entry-editor)))
    (case (table-entry-id entry)
      (:name
       (unless (string= text "")
         (unless (same-name? object text)
           (if (find-object text)
             (warn-user "An object named \"~:(~A~)\" already exists." 
                        text)
             (progn 
               (rename-object object text)
               (map-capella-windows
                #'(lambda (b)
                    (update-window-for-changed-object b object))))))))
      (comment
       (let ((len (length text)))
         (when (and (> (length text) 0) (char= (elt text 0) #\"))
           (setf text (subseq text 1 (1- len))))
         (setf (slot-value object 'comment) text)))
      (t
       ;; normally we want to set a slot to the lisp read item.
       (let ((value (read-from-string text nil 'unset)))
         (if (eq value 'unset)
           (slot-makunbound object (table-entry-id entry))
           (setf (slot-value object (table-entry-id entry)) value))
         (map-capella-windows
          #'(lambda (b) (update-window-for-changed-object b object))
          'browser))))))
