;;; **********************************************************************
;;; Copyright (c) 89-94, 95 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :common-music)

;;;
;;; number streams
;;;

(defclass number-pattern (item-pattern)
  ((from :initform 0 :initarg :from)
   (to   :initarg :to :initarg :below :initarg :downto :initarg :above)
   (by   :initform 1 :initarg :by)))
   
(defmethod initialize-instance :after ((object number-pattern) 
                                       &rest args)
  (let ((from (slot-value object 'from))
        (by (slot-value object 'by))
        (flag 0) 
        to)
    (loop for arg in args by #'cddr
	  while (= flag 0)
          do
      (case arg
        (:below  (setf flag +numbers-below+))
        (:downto (setf flag +numbers-downto+))
        (:above  (setf flag +numbers-above+))))
    (setf to
      (if (slot-boundp object 'to)
          (slot-value object 'to)
          (progn (setf flag  +numbers-unbounded+)
                 (if (floatp from)
                     most-positive-single-float
                     most-positive-fixnum))))
    (setf (slot-value object 'flags)
      (logior (slot-value object 'flags) flag))
    (setf (slot-value object 'data)
      (vector from to by))))

(defmethod reset-period :before ((object number-pattern))
  (let ((resets (slot-value object 'data))
        (flags (slot-value object 'flags))
        (by (slot-value object 'by))
        from to)
    (unless (and (logtest flags +numbers-unbounded+)
                 (slot-boundp object 'to))
      (setf (slot-value object 'from) (item (svref resets 0)))
      (setf (slot-value object 'to)   (item (svref resets 1))))

      (setf (slot-value object 'by)   (item (svref resets 2)))
      (if (or (logtest flags +numbers-downto+) 
              (logtest flags +numbers-above+))
          (setf to (slot-value object 'from) 
                from (slot-value object 'to))
        (setf from (slot-value object 'from) 
              to (slot-value object 'to)))
      ;; numbers compute their period lengths. is this correct?
      (setf (period-length (slot-value object 'period))
        (if (logtest flags +numbers-unbounded+)
            1
            (if (or (logtest flags +numbers-below+)
                    (logtest flags +numbers-above+))
                (ceiling (abs (/ (- to from) by)))
              (multiple-value-bind (int rem) 
                                   (ceiling
                                     (abs (/ (- to from) by)))
                  (if (= rem 0) (1+ int) int)))))))

(defclass linear-number-pattern (number-pattern) ())
(defclass random-number-pattern (number-pattern) 
  ((ran :initform *cm-state* :initarg using :initarg :using)))

(defmethod next-in-pattern ((object linear-number-pattern))
  (let ((flags (slot-value object 'flags))
        (from (slot-value object 'from)))
    (cond ((or (logtest flags +numbers-downto+)
               (logtest flags +numbers-above+)) 
           (decf (slot-value object 'from) (slot-value object 'by)))
          (t
           (incf (slot-value object 'from) (slot-value object 'by))))
    from))

(defmethod next-in-pattern ((object random-number-pattern))
  (let ((flags (slot-value object 'flags)))
    (cond ((logtest flags +numbers-downto+) 
           (random-downto (slot-value object 'from)
                          (slot-value object 'to)
                          (slot-value object 'by)
                          (slot-value object 'ran)))
          ((logtest flags +numbers-below+) 
           (random-below (slot-value object 'from)
                         (slot-value object 'to)
                         (slot-value object 'by)
                         (slot-value object 'ran)))
          ((logtest flags +numbers-above+) 
           (random-above (slot-value object 'from)
                         (slot-value object 'to)
                         (slot-value object 'by)
                         (slot-value object 'ran)))
          (t 
           (random-to (slot-value object 'from)
                      (slot-value object 'to)
                      (slot-value object 'by)
                      (slot-value object 'ran))))))

(defun random-to (from to by state)
  (+ from (* by (random (1+ (round (/ (- to from) by))) state))))

(defun random-above (from to by state)
  (- from (* by (random (round (/ (- from to) by)) state))))

(defun random-below (from to by state)
  (+ from (* by (random (round (/ (- to from) by)) state))))

(defun random-downto (from to by state)
  (- from (* by (random (1+ (round (/ (- from to) by))) state))))

;;;
;;; numbers only supports two pattern types.
;;;

(defclass linear-number-stream (item-stream linear-number-pattern)
  ())

(defclass random-number-stream (item-stream random-number-pattern)
  ())

#|
(read-items (setf x (numbers from 1 to 10)))
(read-items (setf x (numbers below 10)))
(read-items (setf x (numbers downto -2 in random by .5)))
(read-items (setf x (numbers  from 1 to 10 by .1)))
(read-items (setf x (numbers from 0.2 by 0.01)) 10)
|#
         
;;;
;;; note streams
;;;

(defclass scale-mixin ()
  ((scale :initform *standard-scale* :initarg :scale :initarg :of
          :initarg :mode)))

;;;
;;; dont break if stream is passed.
;;;

(defmethod note ((note item-stream) &optional scale)
  (declare (ignore scale))
  note)

(defmethod pitch ((note item-stream) &optional scale)
  (declare (ignore scale))
  note)

(defmethod degree ((note item-stream) &optional scale)
  (declare (ignore scale))
  note)

;;;
;;; note streams return symbolic note name, degree stream return integer
;;; scale degrees, pitch streams return floating point frequencies.
;;;

(defclass note-stream (item-stream scale-mixin)
  ())

(defclass degree-stream (item-stream scale-mixin)
  ())

(defclass pitch-stream (item-stream scale-mixin) 
  ())

;;;
;;; the after methods convert the stream's value to a scale reference.
;;;

(defmethod item :after ((object note-stream))
  ;(declare (optimize (speed 3) (safety 0)))
  (setf (slot-value object 'value)
    (scale-note (slot-value object 'value)
                (slot-value object 'scale))))

(defmethod item :after ((object pitch-stream))
  ;(declare (optimize (speed 3) (safety 0)))
  (setf (slot-value object 'value)
    (scale-pitch (slot-value object 'value)
                 (slot-value object 'scale))))

(defmethod item :after ((object degree-stream))
  ;(declare (optimize (speed 3) (safety 0)))
  (setf (slot-value object 'value)
    (scale-degree (slot-value object 'value)
                  (slot-value object 'scale))))

;;;
;;; note, degree and pitch stream definitions
;;;

(defclass cyclic-note-stream (note-stream cycle-pattern) ())
(defclass cyclic-degree-stream (degree-stream cycle-pattern) ())
(defclass cyclic-pitch-stream (pitch-stream cycle-pattern) ()) 

(defclass palindromic-note-stream (note-stream palindrome-pattern) ())
(defclass palindromic-degree-stream (degree-stream palindrome-pattern) ())
(defclass palindromic-pitch-stream (pitch-stream palindrome-pattern) ()) 

(defclass sequential-note-stream (note-stream sequence-pattern) ())
(defclass sequential-degree-stream (degree-stream sequence-pattern) ())
(defclass sequential-pitch-stream (pitch-stream sequence-pattern) ()) 

(defclass heap-note-stream (note-stream heap-pattern) ())
(defclass heap-degree-stream (degree-stream heap-pattern) ())
(defclass heap-pitch-stream (pitch-stream heap-pattern) ())

(defclass random-note-stream (note-stream random-pattern) ())
(defclass random-degree-stream (degree-stream random-pattern) ())
(defclass random-pitch-stream (pitch-stream random-pattern) ())

(defclass graph-note-stream (note-stream graph-pattern) ())
(defclass graph-degree-stream (degree-stream graph-pattern) ())
(defclass graph-pitch-stream (pitch-stream graph-pattern) ())

(defclass functional-note-stream (note-stream function-pattern) ())
(defclass functional-degree-stream (degree-stream function-pattern) ())
(defclass functional-pitch-stream (pitch-stream function-pattern) ())

(defclass rotational-note-stream (note-stream rotation-pattern) ())
(defclass rotational-degree-stream (degree-stream rotation-pattern) ())
(defclass rotational-pitch-stream (pitch-stream rotation-pattern)  ())

(defclass accumulating-note-stream (note-stream accumulation-pattern) ())
(defclass accumulating-degree-stream (degree-stream accumulation-pattern) ())
(defclass accumulating-pitch-stream (pitch-stream accumulation-pattern) ()) 

(defclass rewrite-note-stream (note-stream rewrite-pattern) ())
(defclass rewrite-degree-stream (degree-stream rewrite-pattern) ())
(defclass rewrite-pitch-stream (pitch-stream rewrite-pattern) ()) 

;;;
;;; chord streams return the state :chording if not at end of period.
;;;

(defclass chord-mixin () ())

(defclass chord-stream (chord-mixin item-stream cycle-pattern)
  ())

(defclass note-chord-stream (chord-mixin note-stream cycle-pattern)
  ())

(defmethod item :around ((stream chord-mixin)) 
  ;(declare (optimize (speed 3)(safety 0)))
  (multiple-value-bind (value state) (call-next-method)
    (values value (or state ':chording))))

;;;
;;; note parsing. 
;;;

(defvar *standard-octave* 4)
(defparameter *respect-note-spelling* #+cmn t #-cmn nil)

(defmethod datum-parser ((object note-stream))
  (let ((scale (slot-value object 'scale)))
    (setf *standard-octave* 4)
    #'(lambda (x )  (parse-note x scale))))

(defmethod datum-parser ((object degree-stream))
  (let ((scale (slot-value object 'scale)) )
    (setf *standard-octave* 4)
    #'(lambda (x) (parse-note x scale))))

(defmethod datum-parser ((object pitch-stream))
  (let ((scale (slot-value object 'scale)))
    (setf *standard-octave* 4)
    #'(lambda (x) (parse-note x scale))))

(defmethod parse-note ((ref integer) (scale essential-scale))
  (if (gethash ref (slot-value scale 'entries))
      ref
    (error "~A is not a degree in ~A." ref scale)))

(defmethod parse-note ((ref float) (scale essential-scale))
  ref)

(defmethod parse-note ((ref item-stream) (scale essential-scale))
  ref)

(defmethod parse-note ((ref expr) (scale essential-scale))
  ref)

(defmethod parse-note (ref (object mode))
  (parse-note ref (slot-value object 'scale)))

(defmethod parse-note ((ref symbol) (scale essential-scale))
  (if (eq ref 'r)
    *rest-name*
    (let ((thing (cdr (gethash ref (slot-value scale 'entries))))
          (divisions (slot-value scale 'divisions-per-octave)))
      (if thing
        (progn
          (setf *standard-octave* 
                (gethash (+ (floor thing divisions) 1024)
                         (slot-value scale 'tokens)))
          (setf thing ref))
        ;; get position of token, add in default octave and return symbol
        ;; at that degree
        (let* ((tokens (slot-value scale 'tokens)) 
               (octave (gethash *standard-octave* tokens)) 
               (interval (gethash ref tokens)) 
               degree entry) 
          (unless (and interval octave)
            (error "~S is not a note in ~S." ref scale))
          (setf degree (degree (+ (* divisions octave) interval)))
          (setf entry (gethash degree (slot-value scale 'entries)))
          (if *respect-note-spelling*
            (let* ((string (symbol-name ref))
                   (length (length string)))
              (setf thing (find string (cdr entry) 
                                :key #'symbol-name
                                :test #'(lambda (x y)
                                          (string= x y :end2 length)))))
            (setf thing (cadr entry)))))
      thing)))

(defmethod parse-note ((ref symbol) (scale general-scale))
  (if (> (slot-value scale 'number-of-octaves) 0)
      (call-next-method)
    (if (scale-degree ref scale) 
        ref
      (error "~S is not a note in ~S." ref scale))))

;;;
;;; interval streams
;;;

(defstruct offset 
  value    ; current offset value
  mode     ; one of :constant, :linked-to, :initially-from, :from
  source)  ; reinitialization source if mode not :constant.

(defclass interval-mixin (scale-mixin)
  ((offset :initform 0 :initarg :offset :initarg :from :initarg :on
           :initarg :linked-to :initarg :initially-from)))

(defclass essential-interval-stream (item-stream interval-mixin)
  ())

(defmethod initialize-instance :after ((object essential-interval-stream) 
                                       &rest args)
  (let ((offset (slot-value object 'offset))
        (mode ':from)
        (value nil)
        (source nil)
        (scale (slot-value object 'scale)))
    (loop for arg in args by #'cddr
          do 
      (case arg
        (:initially-from 
         (setf mode arg 
               value (read-offset offset (slot-value object 'scale))))
        (:linked-to 
         (setf mode arg source offset))
        (:on 
         (setf (slot-value object 'flags)
           (logior (slot-value object 'flags)
             +parallel-offsets+)))))
    (when (typep scale 'mode)
     (setf (slot-value object 'flags)
           (logior (slot-value object 'flags) +modal-intervals+))) 
    (when (eq mode ':from)
      (if (or (numberp offset) (symbolp offset))
          (setf mode ':constant 
                ; source (scale-degree offset scale)
                source (read-offset offset scale)
                )
        (setf source offset)))
    (setf (slot-value object 'offset)
      (make-offset :mode mode :value value :source source))))

(defmethod datum-parser ((object essential-interval-stream))
  #'(lambda (x) 
      (cond ((or (integerp x) (item-stream? x)) x)
            ((restp x) *rest-degree*)
            (t (error "~S not an interval." x)))))

;;;
;;; interval streams return an interval added to an offset.
;;; step streams return an offset postincremented by the interval.
;;;

(defclass interval-stream (essential-interval-stream)
     ())

(defclass step-stream (essential-interval-stream)
     ())

;;;
;;; offsets are calculated once per period for interval and
;;; step streams, voicing streams calculate once per read.
;;;

(defmethod reset-period :after ((object essential-interval-stream))
  (unless (logtest (slot-value object 'flags)
                   +parallel-offsets+)
    (reset-offset (slot-value object 'offset)
                  (slot-value object 'scale))))

(defmethod next-in-pattern :before ((object essential-interval-stream))
    (when (logtest (slot-value object 'flags)
                   +parallel-offsets+)
      (reset-offset (slot-value object 'offset) 
                    (slot-value object 'scale))))

;;;
;;; reset-offset computes a new offset value according to its mode.
;;;

(defun reset-offset (offset scale)
  (let ((source (offset-source offset))
        mode next)
    ;; if no source the offset was specified as :initially-from
    (when source
      (setf mode (offset-mode offset))
      (cond ((eq mode :constant)
             ;; just reset to constant value
             (setf next (offset-source offset)))
            ((eq mode ':linked-to)
             ;; get value from delegate stream.
             (let ((stream (find-item-stream (offset-source offset)))) ;(find-stream )
               (setf next
                 (read-offset (linked-offset stream) scale))))
            (t
             ;; recompute new value from source
             (setf next (read-offset source scale))))
      (setf (offset-value offset) 
        (if (restp next) *rest-degree* next)))))

(defmethod linked-offset ((object step-stream))
  (offset-value (slot-value object 'offset)))

(defmethod linked-offset ((object item-stream))
  (slot-value object 'value))

(defmethod read-offset ((offset integer) (scale essential-scale))
   offset)

(defmethod read-offset ((offset t) (scale essential-scale))
  (scale-degree offset scale))

(defmethod read-offset ((offset scale-mixin) (scale essential-scale))
  (scale-degree (item offset) (slot-value offset 'scale)))

(defmethod read-offset ((offset item-stream) (scale essential-scale))
  (scale-degree (item offset) scale))

(defmethod read-offset ((offset expr) (scale essential-scale))
  (read-offset (item offset) scale))

(defmethod read-offset (offset (mode mode))
   (let ((o (item offset))) 
    (transpose mode o)
    (modeify (scale-degree o (slot-value mode 'scale)) mode)))

;;;
;;; the after methods for interval and voicing streams are identical:
;;; the value of the stream is set to the sum of the current offset
;;; and the current value. step streams return the current offset
;;; and postincrement it by the current value. all three streams
;;; optionally coerce the returned value to scale notes or pitches.
;;; scale calculation would ideally be done by interval-mixin, but 
;;; can't be because it's after method would be called before these
;;; methods are called.
;;;

(defmethod item :after ((object interval-stream))
  ;(declare (optimize (speed 3) (safety 0)))
  (let ((value (slot-value object 'value))
        (offset (slot-value object 'offset)))
    (if (or (eq value *rest-degree*)
            (eq (offset-value offset) *rest-degree*))
        (setf (slot-value object 'value) *rest-degree*)
     (let ((flags (slot-value object 'flags)))
       (declare (fixnum flags))
       (setf (slot-value object 'value)
         (cond ((logtest flags +coerce-to-note+)
                (scale-note (+ value (offset-value offset))
                            (slot-value object 'scale)))
               ((logtest flags +coerce-to-pitch+)
                (scale-pitch (+ value (offset-value offset))
                             (slot-value object 'scale)))
               ((logtest flags +modal-intervals+)
                (scale-degree (+ value (offset-value offset))
                              (slot-value object 'scale)))
               (t
                (+ value (offset-value offset)))))))))

(defmethod item :after ((object step-stream))
  ;(declare (optimize (speed 3) (safety 0)))
  (let* ((value (slot-value object 'value))
         (offset (slot-value object 'offset))
         (current (offset-value offset)))
    (if (eq current *rest-degree*)
        (setf (slot-value object 'value) *rest-degree*))
      (if (not (eq value *rest-degree*))
        (let ((flags (slot-value object 'flags)))
         (declare (fixnum flags))
         (setf (slot-value object 'value)
           (cond ((logtest flags +coerce-to-note+)
                  (scale-note current (slot-value object 'scale)))
                 ((logtest flags +coerce-to-pitch+)
                  (scale-pitch current (slot-value object 'scale)))
                 ((logtest flags +modal-intervals+)
                  (scale-degree current (slot-value object 'scale)))
                 (t current)))
         (incf (offset-value offset) value)))))

;;;
;;; interval stream definitions.
;;;

(defclass cyclic-interval-stream (interval-stream cycle-pattern) ())
(defclass cyclic-step-stream (step-stream cycle-pattern) ())

(defclass palindromic-interval-stream (interval-stream palindrome-pattern) ())
(defclass palindromic-step-stream (step-stream palindrome-pattern) ())

(defclass sequential-interval-stream (interval-stream sequence-pattern) ())
(defclass sequential-step-stream (step-stream sequence-pattern) ())

(defclass accumulating-interval-stream (interval-stream accumulation-pattern)
  ())
(defclass accumulating-step-stream (step-stream accumulation-pattern) ())

(defclass random-interval-stream (interval-stream random-pattern) ())
(defclass random-step-stream (step-stream random-pattern) ())

(defclass heap-interval-stream (interval-stream heap-pattern) ())
(defclass heap-step-stream (step-stream heap-pattern) ())

(defclass graph-interval-stream (interval-stream graph-pattern) ())
(defclass graph-step-stream (step-stream graph-pattern) ())

(defclass functional-interval-stream (interval-stream function-pattern) ())
(defclass functional-step-stream (step-stream function-pattern) ())

(defclass rotational-interval-stream (interval-stream rotation-pattern) ())
(defclass rotational-step-stream (step-stream rotation-pattern) ())

(defclass rewrite-interval-stream (interval-stream rewrite-pattern) ())
(defclass rewrite-step-stream (step-stream rewrite-pattern) ())

;;;
;;; the series stream is a cyclic interval stream that supports 
;;; most types of serial (twelve-tone) manipulations. series holds
;;; a canonical version of the series, which is processed each time
;;; the cyclic data repeats. the processing performed depends on the
;;; values of the form, mudulus, multiple and offset slots, which are
;;; read with item each time to reutunr new values. form is the row
;;; form to use, either prime, inversion ,retrograde or retrograde-
;;; inversion. multiple is a multiplicitive on each value in the series
;;; and modulus removes that interval from the computed value.
;;;


(defclass series-stream (cyclic-interval-stream)
  ((form     :initform nil :initarg forming :initarg :forming)
   (modulus  :initform nil :initarg :modulus)
   (multiple :initform nil :initarg multiple :initarg :multiple)
   (series)))

(defmethod initialize-instance :after ((object series-stream) &rest args)
  (declare (ignore args))
  (setf (slot-value object 'series)
    (copy-list (car (slot-value object 'data)))))

(defmethod next-in-pattern :before ((object series-stream))
  ;; recalculate series when data is reset.
  (let ((data (slot-value object 'data)))
    (when (null (cdr data))
      (let ((set (slot-value object 'series))
            (typ (slot-value object 'form))
            (mul (slot-value object 'multiple))
            (mod (slot-value object 'modulus))
            (inv nil)
            (rev nil))
        (when typ (setf typ (item typ)))
        (when mul (setf mul (item mul)))
        (when mod (setf mod (item mod)))
        (ecase typ
          ((nil p prime :p :prime))
          ((r retrograde :r :retrograde)
           (setf rev t))
          ((i :i inversion :inversion)
           (setf inv t))
          ((ri :ri retrograde-inversion :retrograde-inversion)
           (setf inv t rev t)))
        (loop for i in set
              for tail on (car data)
              do
              (when mul (setf i (* i mul)))
              (when inv (setf i (- i)))
              (when mod
                (setf i (mod i (if (minusp i) (- mod) mod))))
              (setf (car tail) i))
        (when rev
          (setf (car data) (nreverse (car data))))))))

;;;
;;; rhythm streams
;;;

(defparameter *standard-tempo* 60.0 "The default tempo factor.")

(defparameter *clock-mode* ':seconds "The system clock mode.")

(defun in-tempo (tempo &optional (pulse 'q))
  (unless (and (numberp tempo) (> tempo 0))
    (error "Tempo ~A is not greater than 0." tempo))
  (setf *standard-tempo* (* tempo (rhythm pulse 60))))

(defun clock-mode (&optional (mode nil mp))
  (when mp
    (ecase mode
      ((:seconds seconds 1) 
       (setf *clock-mode* ':seconds))
      ((:milliseconds milliseconds 1000)
       (setf *clock-mode* ':milliseconds))))
  *clock-mode*)

(defun seconds (msec) (/ msec 1000.0))

(defun milliseconds (sec) (floor (* sec 1000)))

(defclass rhythm-stream (item-stream)
  ((tempo :initform *standard-tempo*
          :initarg :tempo :initarg :scaler)))

(defmethod initialize-instance :after ((object rhythm-stream) 
                                       &rest args)
  (declare (ignore args))
  (when (numberp (slot-value object 'tempo))
    (setf (slot-value object 'tempo)
          (/ (if (eq *clock-mode* ':seconds) 60.0 60)
             (slot-value object 'tempo)))))

#|
(if (eq *clock-mode* ':seconds)
            (* 4 (/ 60.0 (slot-value object 'tempo)))
            (round (* 4000 (/ 60 (slot-value object 'tempo)))))
|#
(defmethod rhythm ((rhythm t) &optional (tempo *standard-tempo*))
  (if (eq *clock-mode* :seconds)
    (/ (* 4 (/ 60.0 tempo)) (parse-rhythm rhythm)) 
    (round (* 4000 (/ 60 tempo)) (parse-rhythm rhythm))))

(defmethod rhythm ((rhythm rhythm-stream) &optional tempo)
  (declare (ignore tempo))
  rhythm)

(defmethod rhythm ((rhythm expr) &optional tempo)
  (declare (ignore tempo))
  rhythm)

(defmethod datum-parser ((object rhythm-stream))
  #'parse-rhythm)

;;;
;;; a rhythm should only be scaled one time so that substreams return
;;; values in their own local tempi. this around method scales a value
;;; only if it is eq with the stream's actual datum. otherwise the datum
;;; is a substream and the scaling is skipped.  since the datum slot is
;;; accessed, the class precedence list must insure that this around
;;; method is called after the main around method has set a valid
;;; next datum.
;;;

(defmethod item :after ((object rhythm-stream))
  (let ((value (slot-value object 'value)))
    (when (eq value (slot-value object 'datum))
      (setf (slot-value object 'value)
            (if (eq *clock-mode* ':seconds)
              (/ (* 4 (read-tempo (slot-value object 'tempo) value))
                 value)
              (round (* 4000 (read-tempo (slot-value object 'tempo) value))
                     value))))))

#|
(clock-mode 1)
(read-items (setf y (rhythms q for 8 tempo 90)))
(read-items (setf y (rhythms q for 8 tempo (tempo 0 60 6 90 8 120))))
(clock-mode 1000)
(read-items (setf y (rhythms q for 8 tempo 90)))
(read-items (setf y (rhythms q for 8 tempo (tempo 0 60 6 90 8 120))))
|#

;;;
;;; tempo functions support dynamic tempo change. the tempo scaler is updated
;;; either before or after the calculation, depending on whether the change
;;; should start on, or or end in, the final tempo.
;;;

(defclass tempo-function ()
  ((coords :initarg :coords)
   (beat   :initform 'q :initarg :beat)
   (time   :initform 0)
   (mode   :initform ':before :initarg :mode)
   (length :initarg :length :initform nil)))

(defmethod initialize-instance :after ((object tempo-function) &rest args)
  (declare (ignore args))
  (macrolet ((coord-x (l) `(car ,l))
             (coord-y (l) `(car (cdr ,l))))
    ;; convert tempi to scalers and beat length to time.
    (let ((sec (if (eq *clock-mode* ':seconds) 60.0 60)))
      (setf (slot-value object 'beat)
            (rhythm (slot-value object 'beat) sec))
      (loop with last
            for tail on (slot-value object 'coords) by #'cddr
            do
            (setf (coord-y tail) (/ sec (coord-y tail))
                  last tail)  
	        finally
            (let ((length (slot-value object 'length)))
		      (setf (slot-value object 'length)
			        (if length
                        (* length (slot-value object 'beat))
                        (* (coord-x last) 
                           (slot-value object 'beat)))))))))

;;;
;;; read-tempo returns a tempo value.  tempo may be a number, item
;;; stream, envelope or function.
;;;

(defmethod read-tempo ((tempo t) rhythm)
  (declare (ignore rhythm))
  (if (functionp tempo)
      (funcall tempo)
    (error "Bad tempo: ~S." tempo)))

(defmethod read-tempo ((tempo number) rhythm)
  (declare (ignore rhythm))
  tempo)

(defmethod read-tempo ((tempo item-stream) rhythm)
  (declare (ignore rhythm))
  (item tempo))

(defmethod read-tempo ((object tempo-function) rhythm)
  (setf rhythm
        (/ (if (eq *clock-mode* ':seconds) 4.0 4000)rhythm))
  (if (eq (slot-value object 'mode) ':before)
    (interp1 (/ (setf (slot-value object 'time)
                      (+ (mod (slot-value object 'time) 
                              (slot-value object 'length))
                         rhythm))
                (slot-value object 'beat))
             (slot-value object 'coords))
    (let ((val (interp1 (/ (mod (slot-value object 'time)
                                (slot-value object 'length))
                           (slot-value object 'beat))
                        (slot-value object 'coords))))
      (incf (slot-value object 'time) rhythm)
      val)))

;;;
;;; parse-rhythm returns canonical rhythm values
;;;

(defmethod parse-rhythm ((rhythm t))
  (error "No such rhythm: ~S" rhythm))

(defmethod parse-rhythm ((rhythm number))
  rhythm ;(if (= rhythm 0) 0 (/ 4 rhythm))
  )

(defmethod parse-rhythm ((rhythm item-stream))
  rhythm)

(defmethod parse-rhythm ((rhythm symbol))
  (parse-rhythm-string (symbol-name rhythm)))

(defun parse-rhythm-string (rhythm)
  (flet ((parse-rhythm-token (str)
	   (let* ((end (length str))
		  (beg (if (char= (elt str 0) #\T) 1 0))
		  (trp (= beg 1))
		  (chr (char str beg))
		  (pos (position chr '(#\M #\L #\B #\W #\H #\Q #\E #\S)))
		  rhy)
	     (cond (pos
		    (incf beg)
		    (setf rhy (expt 2 (- pos 3))))
		   (t
		    (setf pos beg)
		    (loop with flg = nil
			  while (< beg end)
			  do (setf chr (elt str beg))
			  if (or (digit-char-p chr)
				 (and (char= chr #\.)
				      (not flg)
				      (< beg (1- end))
				      (digit-char-p (elt str (1+ beg)))
				      (setf flg t)))
			  do (incf beg)
			  else do (return))
		    (if (> beg pos)
                      (setf rhy (read-from-string (subseq str pos beg)))
		      (error "Not a rhythm: ~A" str)) 
		    (when (and (<= (setf pos (+ beg 2)) end)
			       (or (string= str "TH" :start1 beg :end1 pos)
				   (string= str "ND" :start1 beg :end1 pos)
				   (string= str "ST" :start1 beg :end1 pos)))
		      (setf beg pos))))
	     (loop while (and (< beg end)
			      (char= (elt str beg) #\.))
		   count (incf beg) into dot
		   finally (unless (= dot 0)
			     (setf rhy (/ rhy (- 2 (expt 2 (- dot)))))))
	     (when trp
	       (setf rhy (* rhy 3/2)))
	     (if (= beg end)
               rhy
	       (error "Can't parse ~A as a rhythm." str))))
	 (next-token-position (string lb len)
	   (loop with chr for i from lb below len
                 do (setf chr (elt string i))
                 until (find chr '(#\+ #\- #\* #\/) :test #'char=)
                 finally (return i))))
    ;; parse rhythmic expression. operator precedence is not supported. 
    (let* ((len (length rhythm))
           (lb 0)
           (ub (next-token-position rhythm lb len)))
      (unless (< lb ub) (error "Not a rhythm: ~A" rhythm))
      (loop with sum = (/ 1 (parse-rhythm-token (subseq rhythm lb ub)))
            and val and op
            for i from 0
            while (< ub len)
            do
            (setf op (elt rhythm ub))
            (setf lb (1+ ub))
            (setf ub (next-token-position rhythm lb len))
            (unless (< lb ub)
              (error "Not a rhythm: ~A." rhythm))
            (cond ((char= op #\+)
                   (setf val (/ 1 (parse-rhythm-token (subseq rhythm lb ub))))
                   (incf sum val))
                  ((char= op #\-)
                   (setf val (/ 1 (parse-rhythm-token (subseq rhythm lb ub))))
                   (decf sum val))
                  ((char= op #\*)
                   (setf val (read-from-string (subseq rhythm lb ub)))
                   (setf sum (* sum val)))
                  ((char= op #\/)
                   ;(error "Rhythm division is not supported.")
                   (setf val (read-from-string (subseq rhythm lb ub)))
                   (setf sum (/ sum val))))
            finally (return (/ 1 sum))))))

;;;
;;; rhythm stream classes
;;;

(defclass cyclic-rhythm-stream (rhythm-stream cycle-pattern) ())
(defclass palindromic-rhythm-stream (rhythm-stream palindrome-pattern) ())
(defclass sequential-rhythm-stream (rhythm-stream sequence-pattern) ())
(defclass accumulating-rhythm-stream (rhythm-stream accumulation-pattern) ())
(defclass heap-rhythm-stream (rhythm-stream heap-pattern) ())
(defclass random-rhythm-stream (rhythm-stream random-pattern) ())
(defclass graph-rhythm-stream (rhythm-stream graph-pattern) ())
(defclass functional-rhythm-stream (rhythm-stream function-pattern) ())
(defclass rotational-rhythm-stream (rhythm-stream rotation-pattern) ())
(defclass rewrite-rhythm-stream (rhythm-stream rewrite-pattern) ())

;;;
;;; ampliltude streams
;;;

;;;
;;; ampitude streams map logical amplitudes to actual amplitude values.
;;; logical amplitudes are numbers between 0 and 1.0 or symbols in
;;; the set *amplitude-symbols*.
;;;

(defparameter *amplitude-minimum* 0.0
  "Minimum amplitude when logical amplitude is 0.0")

(defparameter *amplitude-maximum* 1.0 
  "Maximum amplitude when logical amplitude is 1.0")

(defparameter *amplitude-power* 1 "Amplitude power curve.")

(defprop niente :amplitude-weight 0)
(defprop pppp :amplitude-weight .1)
(defprop ppp :amplitude-weight .2)
(defprop pp :amplitude-weight .3)
(defprop p :amplitude-weight .4)
(defprop mp :amplitude-weight .5)
(defprop mf :amplitude-weight .6)
(defprop f :amplitude-weight .7)
(defprop ff :amplitude-weight .8)
(defprop fff :amplitude-weight .9)
(defprop ffff :amplitude-weight 1)

(defclass amplitude-stream (item-stream) 
  ((power :initform *amplitude-power* :initarg :power)
   (softest :initform *amplitude-minimum* :initarg :softest)
   (loudest :initform *amplitude-maximum* :initarg :loudest)))


(defmethod initialize-instance :after ((object amplitude-stream)
                                       &rest args)
  (declare (ignore args))
  ;; loudest is converted to loudest-softest to avoid recomputation
  ;; each time item is called.
  (setf (slot-value object 'loudest)
    (- (slot-value object 'loudest)
       (slot-value object 'softest))))

(defmethod item :after ((object amplitude-stream))
  (setf (slot-value object 'value)
     (amplitude-value (slot-value object 'loudest)
                      (slot-value object 'value)
                      (slot-value object 'softest) 
                      (slot-value object 'power))))

(defmethod amplitude-value ((range integer) amp min power)
  (values (floor (+ min (* range (expt amp power))))))

(defmethod amplitude-value ((range number) amp min power)
  (+ min (* range (expt amp power))))

(defmethod datum-parser ((object amplitude-stream))
  #'parse-amplitude)

(defmethod parse-amplitude ((amp symbol))
  (or (get amp ':amplitude-weight)
      (error "~S is not an amplitude." amp)))

(defmethod parse-amplitude ((amp number))
  (unless (and (>= amp 0.0) (<= amp 1.0))
    (error "~S is not an amplitude." amp))
  amp)

(defmethod parse-amplitude ((amp item-stream))
  amp)

(defmethod parse-amplitude ((amp expr))
  amp)

(defmethod parse-amplitude ((amp t))
  (error "~S is not an amplitude." amp))

;;;
;;;
;;;

(defclass cyclic-amplitude-stream (amplitude-stream cycle-pattern) ())
(defclass sequential-amplitude-stream (amplitude-stream sequence-pattern) ())
(defclass palindromic-amplitude-stream (amplitude-stream palindrome-pattern)
  ())
(defclass random-amplitude-stream (amplitude-stream random-pattern) ())
(defclass graph-amplitude-stream (amplitude-stream graph-pattern) ())
(defclass heap-amplitude-stream (amplitude-stream heap-pattern) ())
(defclass rotational-amplitude-stream (amplitude-stream rotation-pattern) ())
(defclass functional-amplitude-stream (amplitude-stream function-pattern) ())
(defclass accumulating-amplitude-stream (amplitude-stream accumulation-pattern)
  ())
(defclass rewrite-amplitude-stream (amplitude-stream rewrite-pattern) ())

;;;
;;; user functions
;;;

(defmethod amplitude ((amp t) &key (softest *amplitude-minimum*) min max
		      (loudest *amplitude-maximum*) (power *amplitude-power*))
  (setf min (or min softest))
  (setf max (or max loudest))
  (amplitude-value (- max min) (parse-amplitude amp) min power))

(defmethod amplitude ((amp item-stream) &key min max power softest loudest)
  (declare (ignore min max power loudest softest))
  amp)

;;;
;;; pathname streams (Anders)
;;;

(defvar *pathnames-default-name*
  #+clm clm::default-sound-file-name
  #-clm "test.snd")

(defclass pathname-stream (item-stream)
  ((dir :initform (directory-namestring *pathnames-default-name*)
	:initarg :directory :initarg directory)
   (suffix :initform (pathname-type (pathname *pathnames-default-name*))
	   :initarg type :initarg :type)
   (pathlist :initform nil :initarg pathlist :initarg :pathlist)))

(defclass cyclic-pathname-stream (pathname-stream cycle-pattern) ())
(defclass sequential-pathname-stream (pathname-stream sequence-pattern) ())
(defclass palindromic-pathname-stream (pathname-stream palindrome-pattern) ())
(defclass random-pathname-stream (pathname-stream random-pattern) ())
(defclass graph-pathname-stream (pathname-stream graph-pattern) ())
(defclass heap-pathname-stream (pathname-stream heap-pattern) ())
(defclass rotational-pathname-stream (pathname-stream rotation-pattern) ())
(defclass functional-pathname-stream (pathname-stream function-pattern) ())
(defclass accumulating-pathname-stream (pathname-stream accumulation-pattern) ())
(defclass rewrite-pathname-stream (pathname-stream rewrite-pattern) ())

(defmethod item :after ((object pathname-stream))
  (setf (slot-value object 'value)
	(pathnamestring (slot-value object 'value)
			 (slot-value object 'dir)
			 (slot-value object 'suffix)
			 (slot-value object 'pathlist))))

(defmethod pathnamestring ((value integer) dir suffix pathlist)
  (let ((default (make-pathname :directory dir :type suffix)))
    (namestring (merge-pathnames (nth value pathlist) default))))

(defmethod pathnamestring (value dir suffix pathlist)
  (declare (ignore pathlist))
  (let ((default (make-pathname :directory dir :type suffix)))
    (namestring (merge-pathnames value default))))

(defmethod datum-parser ((object pathname-stream))
  #'parse-pathname)

(defmethod parse-pathname ((name string))
  name)

#-pcl
(defmethod parse-pathname ((name pathname))
  name)

(defmethod parse-pathname ((name item-stream))
  name)

(defmethod parse-pathname ((name expr))
  name)

(defmethod parse-pathname ((name integer))
  (when (< name 0)
	(error "Illegal filename: ~A" name))
  name)

(defmethod parse-pathname ((name t))
  (error "Illegal filename: ~s" name)) 

;;;
;;; multiple items
;;;

(defclass multiple-item () 
  ((items :accessor multiple-item-items :initarg :items)))

(defmethod item-stream? ((object multiple-item)) object)

(defmethod make-multiple-item ((name symbol) &rest args)
  (let ((item (make-instance name)))
    (setf (slot-value item 'items)
      (parse-item-elements item args))
    item))
  
(defmethod parse-item-elements ((item multiple-item) args)
  args)

(defmethod filter-item-elements ((item multiple-item) &rest values)
  values)

(defmacro defmultiple-item-parser (name args &body body)
  (let ((namevar (gensym))
	(argsvar (gensym)))
  `(defmethod parse-item-elements ((, namevar ,name) ,argsvar)
     (apply (function (lambda ,args ,@body))
	    ,argsvar))))

(defmacro defmultiple-item-filter (name args &body body)
  (let ((namevar (gensym))
	(argsvar (gensym)))
  `(defmethod filter-item-elements ((,namevar ,name) &rest ,argsvar)
     (apply (function (lambda ,args ,@body))
	    ,argsvar))))

(defmacro multiple-item-bind (vars item &body body)
  `(multiple-value-bind ,vars (values-list ,item)
     ,@body))

#|
defmultiple-item
The :element-period option has now the following syntax:

	(:element-period :ANY &rest elements)
	means any element in elements reaches end of period 
        returns end of period

	(:element-period :ALL &rest elements)
	means end of period when all elements simultaneously agree.

        (:element-period element)
	the same as (:element-period :all element)
If there is only a single element being tracked, :all or :any have the same effect.
For example:
(defmultiple-item klangs 
                  (pitch rhythm accent amp gesture)
  (:element-period :all))

means: "period ends when both rhythm and accent reach end of period at the same time"
|#

(defmacro defmultiple-item (name elements &rest options)
  (flet ((define-item-method (name descriptors period-type 
                                   state-elements filter)
           (let ((param-var (gensym))
                 (items-var (gensym))
                 (state-var (gensym))
                 (x (gensym))
                 (y (gensym))
                 forms)
             (setf forms
               (loop for d in descriptors
                     for s = (if (find (first d) state-elements) y nil)
                     collect 
                 (cond ((and (eq (first period-type) ':all)
                             (or (not (cdr period-type))
                                 (find (first d) period-type)))
                         `(multiple-value-bind (,x ,y) (item (pop ,items-var))
                            (unless (eq ,y +end-of-stream-token+)
                              (when (or (eq ,state-var nil)
                                        (eq ,state-var +end-of-stream-token+))
                                (setf ,state-var ,y)))
                            ,x))
                       ((and (eq (first period-type) ':any)
                             (or (not (cdr period-type))
                                 (find (first d) period-type)))
                         `(multiple-value-bind (,x ,y) (item (pop ,items-var))
                            (if (eq ,y +end-of-stream-token+)
                                (setf ,state-var +end-of-stream-token+)
                              (unless ,state-var (setf ,state-var ,s)))
                            ,x))
                       (t 
                        `(item (pop ,items-var))))))
             `(defmethod item ((,param-var ,name))
                (let ((,items-var (slot-value ,param-var 'items))
                      (,state-var ,(if (eq (first period-type) ':all)
                                       '+end-of-stream-token+
                                     nil)))
                  ,(if filter
                        `(values (filter-item-elements ,param-var ., forms) 
                                 , state-var)
                    `(values (list .,forms) ,state-var))))))
	 (parse-elements (elements)	; dummy parser for now
	   (loop for e in elements collect (if (consp e) e (list e))))
	 (parse-lambda (option)
	   (let ((lambda? (cadr option)))
	     (when (cddr option)
	       (error "Found extraneous garbage in ~S" option))
	     (unless (and (listp lambda?)
			  (eq (car lambda?) 'lambda)
			  (listp (cadr lambda?))
			  (cddr lambda?))
	       (error "~S is not a well formed lambda body." lambda?))
	     lambda?)))
    (let ((descriptors (parse-elements elements))
	  (period nil) (parser nil) (filter nil) (state nil))
      (dolist (option options)
	(unless (consp option)
	  (error "Option ~S not a list." option))
	(macrolet ((without-duplicates ((var clause) &body body)
		     `(if ,var
			  (error "Found duplicate option: " ,clause)
			(progn ,@body))))
	  (ecase (car option)
	    (:element-period
	     (without-duplicates (period option)
	       (cond ((and (= (length (cdr option)) 1)
                              (find (second option) descriptors 
                                    :test #'member))
		      (setf period (cons ':all (cdr option))))
                     ((and (member (cadr option) '(:all :any))
			   (loop for x in (cddr option) 
                                 always (find x descriptors :test #'member)))
		      (setf period (cdr option)))
		     (t 
		      (error "Bogus :element-period ~S." (cdr option))))))
	    (:element-parser
	     (without-duplicates (parser option)
	       (setf parser (parse-lambda option))) )
	    (:element-state
	     (without-duplicates (state option)
               (if (find (second state) descriptors :test #'member)
                   (setf state (cdr option))
                 (error "Bogus :element-state ~S. Not in ~S" 
                        (cdr option) elements))))
	    (:element-filter
	     (without-duplicates (filter option)
	       (setf filter (parse-lambda option)))))))

      (when (and (not state)
                 (= (length (cdr period)) 1)
                 (find (second period) descriptors :test #'member))
        (setf state (cdr period)))

    `(progn
       (defprop ,name :item-expand t)
       (defclass ,name (multiple-item) ())
       (defmacro ,name (&rest items)
	 `(make-multiple-item ',',name
			      .,(mapcar #'(lambda (x)
					    (if (listp x)
						x
					      (quote-if-necessary x)))
					items)))
       ,(define-item-method name descriptors (or period '(:all)) state filter)
       ,@(if parser 
	     `((defmultiple-item-parser ,name ,(cadr parser) ,@(cddr parser)))
	   nil)
       ,@(if filter
	     `((defmultiple-item-filter ,name ,(cadr filter) 
                                        ,@(cddr filter))))
       (find-class ',name)))))

;;;
;;; macro constructors
;;;

;;;
;;; item stream classes are indexed by their data and pattern types.
;;;

(defprop items :class-index 0)
(defprop notes :class-index 1)
(defprop pitches :class-index 2)
(defprop degrees :class-index 3)
(defprop intervals :class-index 4)
(defprop steps :class-index 5)
(defprop rhythms :class-index 6)
(defprop amplitudes :class-index 7)
(defprop series :class-index 8)
(defprop numbers :class-index 9)
(defprop pathnames :class-index 10)

(defprop cycle :class-index 0)
(defprop random :class-index 1)
(defprop heap :class-index 2)
(defprop palindrome :class-index 3)
(defprop graph  :class-index 4)
(defprop sequence :class-index 5)
(defprop function :class-index 6)
(defprop rotation :class-index 7)
(defprop accumulation :class-index 8)
(defprop rewrite :class-index 9)

;;;
;;; constructors macroexpand any data form that is consp with a symbol in the
;;; car with an :item-expand property of t. otherwise the form is treated as
;;; basic data and not macroexpanded.
;;;

(defprop items :item-expand t)
(defprop notes :item-expand t)
(defprop pitches :item-expand t)
(defprop degrees :item-expand t)
(defprop rhythms :item-expand t)
(defprop amplitudes :item-expand t)
(defprop intervals :item-expand t)
(defprop steps :item-expand t)
(defprop series :item-expand t)
(defprop voicings :item-expand t)  ; remove next release
(defprop numbers :item-expand t)
(defprop pathnames :item-expand t)

(defprop function :disallow-items t)
(defprop numbers  :disallow-items t)

;;;
;;; constructor options are symbols whose :item-option property holds a
;;; parsing list: (keyword quoted? patterns datatypes). keyword is the
;;; option's make-instance keyword argument, quoted? is a flag that, if
;;; true, causes the option value to be quoted during macroexpansion.
;;; patterns and datatypes are the lists of patterns and datatypes that
;;; accept the option, or t if the option applies to all. 
;;;

(defprop ABOVE
  :item-option '(:above nil (cycle random) (numbers)))
(defprop ARGS 
  :item-option '(:args nil (function) t))
(defprop BELOW
  :item-option '(:below nil (cycle random) (numbers)))
(defprop BY
  :item-option '(:by nil (cycle random) (numbers)))
(defprop CHANGE 
  :item-option '(:change nil (rotation) t))
(defprop COUNTING 
  :item-option '(:flags nil t t))

(defprop DIRECTORY
  :item-option '(:directory nil t (pathnames)))
(defprop DOWNTO
  :item-option '(:downto nil (cycle random)(numbers)))
(defprop FOR 
  :item-option '(:for nil t t))
(defprop ELIDING :item-option
  '(:flags nil (palindrome) t))
(defprop ELIDED :item-option
  '(:flags nil (palindrome) t))
(defprop forming 
  :item-option '(:forming nil (cycle) (series)))
(defprop FROM 
  :item-option '(:from nil t (intervals steps series numbers)))
(defprop GENERATIONS
  :item-option '(:generations nil (rewrite) t))
(defprop IN :item-option t)           ; system option not passed on
(defprop INITIALLY
  :item-option '(:initially nil (rewrite) t))
(defprop INITIALLY-FROM
  :item-option '(:initially-from nil t (intervals steps series)))
(defprop LAST-ITEM
  :item-option '(:last-item nil t t))
(defprop LAST-STATE
  :item-option '(:last-state nil t t))
(defprop LINKED-TO 
  :item-option '(:linked-to nil t (intervals steps series)))
(defprop LOUDEST
  :item-option '(:loudest nil t (amplitudes)))
(defprop MODULUS
   :item-option  '(:modulus nil (cycle) (series)))
(defprop MULTIPLE
  :item-option '(:multiple nil (cycle) (series)))
(defprop NAMED
  :item-option '(:name nil t t))
(defprop OF
  :item-option '(:of nil t (notes pitches degrees intervals steps series)))
(defprop ON
  :item-option '(:on nil t (intervals steps series)))
(defprop POWER
  :item-option '(:power nil t (amplitudes)))
(defprop PREVIOUS
  :item-option '(:last t (graph) t))
(defprop PROPS
  :item-option '(:props nil (graph) t))
(defprop RETURNING
  :item-option '(:flags nil t '(intervals steps series)))
(defprop RULES
  :item-option '(:rules nil (rewrite) t))
(defprop SOFTEST
  :item-option '(:softest nil t (amplitudes)))
(defprop PATHLIST
  :item-option '(:pathlist nil t (pathnames)))
(defprop TEMPO
  :item-option '(:tempo nil t (rhythms)))
(defprop TO
  :item-option '(:to nil (cycle random) (numbers)))
(defprop TRAVERSING 
  :item-option '(:flags nil t t))
(defprop TYPE
  :item-option '(:type nil t (pathnames)))
(defprop USING
  :item-option '(:using nil (heap random) t))
(defprop WITH 
  :item-option '(:function nil (graph function) t))

;;;
;;; expand-constructor expands item stream macros into calls to make-instance
;;; with the data selectively evaluated and options mapped keyword arguments.
;;;

(defun expand-constructor (datatype forced-pattern forms env)
  (let ((flags 0) data options class pattern)
    (setf data
      (loop while forms
            for thing = (car forms)
            until (and (symbolp thing) (get thing ':item-option))
            collect (pop forms)))
    (when (and data (get datatype ':items-disallowed))
      (error "~A does not allow explicit items: ~S" datatype data))
    (loop with saved = forms
          while forms
          for option = (pop forms)
          for parser = (or (and (symbolp option)
                                (get option ':item-option))
                           (error "~S is not a constructor option."
                                  option))
          for value = (if forms (pop forms)
                          (error "Malformed (uneven) options list: ~S" 
                                 saved))
          do (cond ((eq option 'in) (setf pattern value))
                   ((eq (first parser) ':flags)
                    (setf flags (logior flags (parse-flag option value))))
                   (t
                    (push value options)
                    (push option options))))
    (if pattern
       (unless (or (null forced-pattern) (eq forced-pattern pattern))
         (error "~S pattern must be ~A." datatype forced-pattern))
      (setf pattern 'cycle))
    (setf class (lookup-class datatype pattern))
    (if data
        (if (or (get pattern ':disallow-items) (get datatype ':disallow-items))
            (error "~A in ~A does not allow explicit items: ~S" 
                   datatype pattern data)
          (setf data (list ':items (relist-for-eval data env pattern))))
      (unless (or (get pattern ':disallow-items) 
                  (get datatype ':disallow-items))
        (error "Missing items for ~A in ~A." datatype pattern)))
    ;; destructively substitute keywords for option names and quote
    ;; values when necessary.
    (loop for tail on options by #'cddr
          for option = (car tail)
          for info = (or (get option ':item-option)
                         (error "~S is not an item stream option."
                                option))
          do 
          (unless (or (eq (third info) t) (find pattern (third info)))
            (error "Option ~S not applicable to ~S." option pattern))
          (unless (or (eq (fourth info) t) (find datatype (fourth info)))
            (error "Option ~S not applicable to ~S." option datatype))
          (setf (car tail) (first info))
          when (second info)
          do (setf (cadr tail) `(quote ,(cadr tail))))
    (when (> flags 0) 
      (push flags options)
      (push ':flags options))
    ;; cobble up the make-instance form with data rewritten for selective 
    ;; evaluation: non-constant atoms are quoted and (...) is replaced by
    ;; (LIST ...) unless its car is a symbol with :item-expand property t.
    `(make-instance ',class ,.data ,.options)))

#|
(defun relist-for-eval (form env)
  (cond ((null form) nil)
        ((listp form) 
         (if (and (symbolp (car form)) (get (car form) ':item-expand))
             (macroexpand form env)
          `(list ,@(loop for f in form 
                         collect (relist-for-eval f env)))))
        ((constantp form) form)
        (t `(quote ,form))))
|#

(defun relist-for-eval (form env &optional pattern)
  (cond ((null form) nil)
        ((listp form) 
         (if (and (symbolp (car form)) (get (car form) ':item-expand))
             (macroexpand form env)
           (if (eq pattern 'random)
               `(list ,.(mapcar #'(lambda (x) (relist-random-datum x env))
                               form))
            `(list ,@(loop for f in form 
                           collect (relist-for-eval f env))))))
        ((constantp form) form)
        (t `(quote ,form))))

(defun relist-random-datum (form env)
  (if (listp form)
      (if (and (symbolp (car form)) (get (car form) ':item-expand))
          (macroexpand form env)
        `(list ,(relist-for-eval (pop form) env)
               ,.(loop for (opt val) on form by #'cddr
                       collect `(quote ,opt)
                       collect val)))
    (if (constantp form) form `(quote ,form))))
           
(defvar *flag-values*
  `((returning
     (,+coerce-to-note+ note) (,+coerce-to-pitch+ pitch) (0 degree))
    (counting
     (,+count-values+ values) (0 periods))
    (traversing
     (,+breadth-first+ breadth) (0 depth))
    (elided
     (,+elide-left+ left start first) (,+elide-right+ right end last)
     (,(logior +elide-left+ +elide-right+) yes t) (0 no nil) )
    (eliding	; duplicate...
     (,+elide-left+ left start first) (,+elide-right+ right end last)
     (,(logior +elide-left+ +elide-right+) yes t) (0 no nil) )
     ))

(defun parse-flag (option value)
  (let ((set (cdr (assoc option *flag-values*)))
        val)
    (unless (setf val (find value set :key #'cdr :test #'member))
      (error "~S value ~S not in ~S" option value
             (loop for x in (cdr set) append (cdr x))))
    (first val)))

(defvar *item-stream-classes* 
  (make-array '(11 10)
    :initial-contents
    '((cyclic-item-stream random-item-stream heap-item-stream
       palindromic-item-stream graph-item-stream sequential-item-stream
       functional-item-stream rotational-item-stream accumulating-item-stream
       rewrite-item-stream)
      (cyclic-note-stream random-note-stream heap-note-stream
       palindromic-note-stream graph-note-stream sequential-note-stream
       functional-note-stream rotational-note-stream accumulating-note-stream
       rewrite-note-stream)
      (cyclic-pitch-stream random-pitch-stream heap-pitch-stream
       palindromic-pitch-stream graph-pitch-stream sequential-pitch-stream
       functional-pitch-stream rotational-pitch-stream 
       accumulating-pitch-stream
       rewrite-pitch-stream)
      (cyclic-degree-stream random-degree-stream heap-degree-stream
       palindromic-degree-stream graph-degree-stream sequential-degree-stream
       functional-degree-stream rotational-degree-stream 
       accumulating-degree-stream
       rewrite-degree-stream)
      (cyclic-interval-stream random-interval-stream heap-interval-stream
       palindromic-interval-stream graph-interval-stream 
       sequential-interval-stream functional-interval-stream
       rotational-interval-stream  accumulating-interval-stream
       rewrite-interval-stream)
      (cyclic-step-stream random-step-stream heap-step-stream
       palindromic-step-stream graph-step-stream sequential-step-stream
       functional-step-stream rotational-step-stream accumulating-step-stream
       rewrite-step-stream)
      (cyclic-rhythm-stream random-rhythm-stream heap-rhythm-stream
       palindromic-rhythm-stream graph-rhythm-stream sequential-rhythm-stream
       functional-rhythm-stream rotational-rhythm-stream 
       accumulating-rhythm-stream
       rewrite-rhythm-stream)
      (cyclic-amplitude-stream random-amplitude-stream heap-amplitude-stream
       palindromic-amplitude-stream graph-amplitude-stream
       sequential-amplitude-stream functional-amplitude-stream
       rotational-amplitude-stream accumulating-amplitude-stream
       rewrite-amplitude-stream)
      (series-stream nil nil nil nil nil nil nil nil nil)
      (linear-number-stream random-number-stream
       nil nil nil nil nil nil nil nil)
      (cyclic-pathname-stream random-pathname-stream heap-pathname-stream
       palindromic-pathname-stream graph-pathname-stream
       sequential-pathname-stream functional-pathname-stream
       rotational-pathname-stream accumulating-pathname-stream
       rewrite-pathname-stream))))

(defun lookup-class (datatype pattern)
  ;; return the item stream class for the two symbolic indicies
  (let ((row (or (get datatype ':class-index)
                 (error "~S is not an item stream type." datatype)))
        (col (or (get pattern ':class-index)
                 (error "~S is not an item pattern type." pattern))))
    (or (aref *item-stream-classes* row col)
        (error "No item stream for ~S and ~S!" datatype pattern))))
;;;
;;;
;;;

(defmacro items (&body body &environment env)
  (expand-constructor 'items nil body env))

(defmacro notes (&body body &environment env)
  (expand-constructor 'notes nil body env))

(defmacro pitches (&body body &environment env)
  (expand-constructor 'pitches nil body env))
            
(defmacro degrees (&body body &environment env)
  (expand-constructor 'degrees nil body env))

(defmacro intervals (&body body &environment env)
  (expand-constructor 'intervals nil body env))
            
(defmacro steps (&body body &environment env)
  (expand-constructor 'steps nil body env))

(defmacro rhythms (&body body &environment env)
  (expand-constructor 'rhythms nil body env))

(defmacro amplitudes (&body body &environment env)
  (expand-constructor 'amplitudes nil body env))

(defmacro series (&body body &environment env)
  (expand-constructor 'series 'cycle body env))

(defmacro numbers (&body body &environment env)
  (expand-constructor 'numbers nil body env))

(defmacro voicings (&body body)
  (declare (ignore body))
  (error "VOICINGS is no longer necessary and has been removed. ~
          Use the INTERVALS macro with the ON option instead. ~
          (VOICINGS ... FROM x) = (INTERVALS ... ON x)"))

(defmacro pathnames (&body body &environment env)
  (expand-constructor 'pathnames nil body env))

(defun make-item-stream (datatype pattern items &rest args)
  (apply #'make-instance (lookup-class datatype pattern) 
        :items items args))

;;;
;;; chord (also the [] read macro)
;;;

(defprop chord :item-expand t)

(defmacro chord (&body body &environment env)
  `(make-chord ,(relist-for-eval body env)))

(defun make-chord (items)
  (if (find-if #'symbolp items)
      (make-instance 'note-chord-stream :items items)
    (make-instance 'chord-stream :items items)))

;;;
;;; crescendo and diminuendo
;;;

(defprop crescendo :item-expand t)
(defprop diminuendo :item-expand t)
(defprop decrescendo :item-expand t)

(defmacro crescendo (&rest body)
  `(make-dynamics ,(relist-for-eval body nil)))

(defmacro diminuendo (&rest body)
  `(make-dynamics ,(relist-for-eval body nil)))
    
(defmacro decrescendo (&rest body)
  (warn "DECRESCENDO works but has been renamed DIMINUENDO")
  `(diminuendo ,@body))

(defun make-dynamics (args)
  (let (from to in inc)
    (loop for (arg val) on args by #'cddr
          do 
      (case arg
        (from (setf from (rationalize (parse-amplitude val))))
        (to (setf to (rationalize (parse-amplitude val))))
        ((in for) (setf in val))))
    (unless from (error "Missing FROM starting amplitude.~A"))
    (unless to (error "Missing TO ending amplitude."))
    (unless in (error "Missing IN stepping value."))
    (setf inc (/ (- to from) (1- in)))
    (make-instance 'cyclic-amplitude-stream
      :items (loop for i below in collect from do (incf from inc))
      :for in)))

;;;
;;;  retrograde returns the reverse of a period
;;;  mirror returns a period followed by its retrograde
;;;

(defprop retrograde :item-expand t)
(defprop mirror :item-expand t)

#|
(defmacro retrograde (x)
  (let ((v (gensym)))
   `(items in function 
           with (let ((,v ,x))
                   #'(lambda () (nreverse (read-items ,v)))))))
|#

;;; fixed retrograde to handle :chording states

(defmacro retrograde (x)
  (let ((v (gensym)))
    `(items in function 
            with 
            (let ((,v ,x))
              #'(lambda ()
                  (multiple-value-bind (l1 l2) (read-items ,v t t)
                    (let (c1 c2 r1 r2)
                      (loop for i1 in l1 as i2 in l2
                            do 
                            (if (eq i2 +eop+)
                              (setf i2 nil)
                              (if (and (null r2) (null i2))
                                (setf i2 +eop+)))
                            (cond ((eq i2 ':chording)
                                   (push i1 c1)
                                   (push i2 c2))
                                  (c1 
                                   (push i1 r1) 
                                   (push i2 r2) 
                                   (loop for x1 in c1 as x2 in c2 
                                         do (push x1 r1) (push x2 r2))
                                   (setf c1 nil c2 nil))
                                  (t 
                                   (push i1 r1)
                                   (push i2 r2))))
                      (values r1 r2))))))))


(defmacro mirror (stream &rest rest)
  `(items in function with
     (let ((items '()) 
           (states '()))
       #'(lambda (s)
           (values
             (if items
                 (prog1 (nreverse items)
                        (setf items nil))
                 (multiple-value-setq (items states) (read-items s t t)))
             states)))
     args (list ,stream) ,@rest))

;;;
;;; repeat repeats the current period of a stream n times. n can 
;;; be an item stream.
;;;

#+lispworks3.1
(eval-when (:load-toplevel
	    :compile-toplevel
	    :execute)
  (shadow 'common-lisp:repeat))

(defprop repeat :item-expand t)

(defmacro repeat (stream n &rest rest)
  `(items in function with
      (let ((items '())
            (states '())
            (limit most-positive-fixnum)
            (count 0))
         #'(lambda (s rep)
             (unless (< count limit)
               (setf items nil count 0))
             (values
               (if items
                   (prog1 items (incf count))
                   (prog1
                     (multiple-value-setq (items states) 
                       (read-items s t t)) 
                     (setf limit (item rep))))
               states)))
      args (list ,stream ,n) ,@rest))

;;;
;;;  motive (also the read macro #@ )
;;;

(defprop motive :item-expand t)

(defmacro motive (name)
  `(find-item-stream ',name))    

;;;
;;; tempo
;;;

(defprop tempo :item-expand t)

(defmacro tempo (&body body)
  (let ((len (length body)))
    (declare (special *coordinates-are-x-y-pairs*))
    (if (< len 3)
      (if (= len 1)
        (car body)    ; number
        (* (car body) (rhythm (cadr body) 60)))			
      (let ((coords
             (loop for tail on body while (numberp (car tail))
                   collect (pop body)))
            (args ())
            (old body)
            from to in)
        (loop with prop and value
              while body
              do
              (setf prop (pop body))
              (unless body
                (error "Malformed (uneven) tempo specification: ~S." 
                       old))
              (setf value (pop body))
              (ecase prop
                (from (setf from value))
                (to (setf to value))
                (in (setf in value))
                (update
                 (push (quote-if-necessary value) args)
                 (push ':mode args))
                ((pulse beat)
                 (push (quote-if-necessary value) args)
                 (push ':beat args))))
        (when (or from to in)
	  (unless from
	    (error "Missing starting tempo: ~S." old))
	  (unless to
	    (error "Missing ending tempo: ~S." old))
	  (unless in
	    (error "Missing number of beats: ~S." old))
	  (setf coords
                (if *coordinates-are-x-y-pairs*
                  (list 0 from in to)
                  (list from 0 to in))))		   
        
        `(make-instance 'tempo-function :coords (list ,.coords) ,.args)))))

;;;
;;; idsel make a selection table used in for markov proceeses (graph stream)
;;;

(defprop idsel :item-expand t)

(defmacro idsel (&rest args &environment env)
  `(list :idsel ., (loop for (ids val) on args by #'cddr
                         collect `(quote ,(if (consp ids) ids (list ids)))
                         collect (relist-for-eval val env))))

;;;
;;; changes constucts one or more rotational changes.
;;;

(defprop changes :item-expand t)

(defmacro changes (&rest args)
  (let ((parsed '()))
    (loop with tail = args
          while tail
          for arg = (pop tail)
          do 
          (push (if tail (pop tail)
                  (error "Malformed (uneven) options list: ~S" args))
                parsed)
          (ecase arg
           ((start :start) (push ':start parsed))
           ((end :end) (push ':end parsed))
           ((step :step) (push ':step parsed))
           ((width :width) (push ':width parsed))))
    `(make-change ,.parsed)))

(defun make-change (&rest args &key start end step width)
  (flet ((chng (start step width end)
                (if end
                  (list (or start 0) (or step 1) (or width 1) end)
                  (if width
                    (list (or start 0) (or step 1) width)
                    (if step 
                      (list (or start 0) step)
                      start)))))
    (macrolet ((next (x) 
               `(if (consp ,x) 
                    (if (cdr ,x) (pop ,x) (car ,x)) 
                  ,x)))
      (if (or (consp start) (consp end) (consp step) (consp width))
          (let ((max (loop for x in args when (consp x) maximize (length x))))
            (make-instance 'cyclic-item-stream
               :items
               (loop repeat max
                     collect (chng (next start) (next step)
                                   (next width) (next end)))))
        (chng start step width end)))))

;;;
;;; expr and fn
;;;

(defprop expr :item-expand t)
(defprop fn   :item-expand t) 

(defmacro expr (form)
  `(make-instance 'expr :function (function (lambda () ,form))))

(defmacro fn (form)
  `(make-instance 'expr :function ,form))
